/* CC-copy.cpp
 *   This program opens the Minecraft window (Minecraft should be
 *   running with a Computercraft terminal open) and creates the files
 *   passed by arguments on the CC computer.
 *
 *   author: Orwell (this_is_1984)
 *   date: October 20 2012
 */

#define DEF_WINDOW_NAME "Minecraft"
#define DEF_STARTUP_TIME 1000
#define DEF_KEY_SPACING 5
#define DEF_TAB_SIZE 2
#define DEF_VERSION 1.41
 
#include <cstdlib>
#include <list>
#include <istream>
#include <fstream>
#include <iostream>
#include <string>
#include <windows.h>

bool gotoMinecraft(const std::string windowName, const unsigned int startupTime)
{
    HWND mc_hwnd = FindWindow(NULL, TEXT( windowName.c_str() ));
    bool res = SetForegroundWindow(mc_hwnd);
    Sleep( startupTime );
    return res;
}

void sendKeyDown(const BYTE vk)
{
    keybd_event(vk, 0x9e, 0, 0);
}

void sendKeyUp(const BYTE vk)
{
    keybd_event(vk, 0x9e, KEYEVENTF_KEYUP, 0);
}

void sendKey(const BYTE vk)
{
    sendKeyDown(vk);
    sendKeyUp(vk);
}

SHORT charToKeycode(const TCHAR c)
{
    return VkKeyScan(c);
}

/*
 * Directly from: http://msdn.microsoft.com/en-us/library/windows/desktop/ms646329(v=vs.85).aspx .
 */
class Key
{
public:
    Key(char c);
    virtual ~Key();
    char getCharacter() { return character; }
    BYTE getKeycode() { return keycode; }
    bool getAlt() { return alt; }
    bool getCtrl() { return ctrl; }
    bool getShift() { return shift; }
private:
    char character;
    BYTE keycode;
    bool shift;
    bool ctrl;
    bool alt;
};

Key::Key(char c)
{
    character = c;
    SHORT code = charToKeycode(c);
    keycode = code & 0x00ff;
    BYTE ms = (code & 0xff00) >> 8;
    shift = ms & 1;
    ctrl = ms & 2;
    alt = ms & 4;         
}

Key::~Key()
{
}



void sendChar(const char c, const unsigned int interval, const bool isOld)
{
    static unsigned int spaceCount = 0;
    static bool isBeginning = true;
    
    Key key(c);
    if(key.getCtrl() && key.getKeycode() != VK_RETURN) {
        sendKeyDown(VK_CONTROL);
        if(key.getAlt()) {
            sendKeyUp(VK_CONTROL);
            Sleep( interval );
            sendKeyDown(VK_CONTROL);
            sendKeyDown(VK_MENU);
        }
    } else if(key.getAlt() && key.getKeycode() != VK_RETURN) {
        sendKeyDown(VK_MENU);
    }
    if(key.getShift() && key.getKeycode() != VK_RETURN) {
        sendKeyDown(VK_SHIFT);
    }

    if (isOld == false) {
        if((key.getKeycode() == VK_SPACE) && isBeginning) {
            spaceCount++;
        } else if((key.getKeycode() == VK_TAB) && isBeginning) {
            spaceCount += 2;
        } else if (isBeginning) {
            isBeginning = false;
        }
    }
    
    if(key.getKeycode() == VK_TAB) {
        sendKey(VK_SPACE);
        Sleep(interval);
        sendKey(VK_SPACE);
    } else {
        sendKey(key.getKeycode());
    }
    
    if(key.getCharacter() == '^') {
        Sleep(interval);
        sendKey(VK_SPACE);
    }
 
    if (isOld == false) {   
        if(key.getKeycode() == VK_RETURN) {
            for(unsigned int i=0; i<spaceCount; i++) {
                     sendKey(VK_BACK);
                 if (i!=(spaceCount-1)) {
                     Sleep(interval);
                 }
             }
             spaceCount = 0;
             isBeginning = true;
         }
    }
    
    if(key.getCtrl() && key.getKeycode() != VK_RETURN) {
        sendKeyUp(VK_CONTROL);
        if(key.getAlt()) {
            sendKeyUp(VK_MENU);
        }
    } else if (key.getAlt() && key.getKeycode() != VK_RETURN) {
        sendKeyUp(VK_MENU);
    }
    
    if(key.getShift() && key.getKeycode() != VK_RETURN) {
        sendKeyUp(VK_SHIFT);
    }
}

void sendString(const std::string s, const unsigned int interval, const bool isOld)
{
    for(unsigned int i=0; i < s.length(); i++) {
        sendChar(s[i], interval, isOld);
    }
}

void startEditingFile(const char* filename, const unsigned int interval, const bool isOld) {
    sendString("rm " + std::string(filename), interval, isOld);
    sendKey(VK_RETURN);
    sendString("edit " + std::string(filename), interval, isOld);
    sendKey(VK_RETURN);
}

void stopEditingFile() {
    sendKey(VK_CONTROL);
    sendKey(VK_RETURN);
    sendKey(VK_CONTROL);
    sendKey(VK_RIGHT);
    sendKey(VK_RETURN);
}

void enterContent(std::istream& file, const unsigned int interval, const bool isOld)
{
    char* ptr_c = new char(1);
    while (!file.eof()) {
        file.read(ptr_c, 1);
        sendChar(*ptr_c, interval, isOld);
        Sleep( interval );
    }
    delete ptr_c;
}

void writeFile(const char* filename, unsigned int interval, const bool isOld) {
    std::ifstream file;
    file.open(filename);
    if(file.is_open()) {
        std::cout << "Writing file '" << filename << "'..." << std::endl;
        startEditingFile(filename, interval, isOld);
        enterContent(file, interval, isOld);
        stopEditingFile();
        file.close();
    } else {
        std::cerr << "Couldn't open file '" << filename << "'." << std::endl;
    }
}

void printUsage()
{
    std::cout << std::endl << "Made by Orwell (this_is_1984)." << std::endl << std::endl;
    std::cout << "This program opens the Minecraft window (Minecraft should be running with a" << std::endl
        << "Computercraft terminal open) and creates the files that are passed by arguments" << std::endl
        << "on the CC computer." << std::endl << std::endl;
    std::cout << "  usage: CC-copy [-w <window>] [-t <interval>] [-s <startup>] [--old] file1 file2 ..." << std::endl;
    std::cout << "    <window> is the title of the Minecraft window." << std::endl;
    std::cout << "              (Typically 'Minecraft' or 'Tekkit')" << std::endl;
    std::cout << "    <interval> is the time between key presses in milliseconds." << std::endl;
    std::cout << "    <startup> is the time that the program waits after switching" << std::endl;
    std::cout << "               to the Minecraft window (in milliseconds)." << std::endl;
}

int main(int argc, char *argv[])
{
    if(argc <= 1) {
        printUsage();
        return 0;
    }
    
    std::string windowName = std::string(DEF_WINDOW_NAME);
    unsigned int interval = DEF_KEY_SPACING;
    unsigned int startupTime = DEF_STARTUP_TIME;
    bool isOld = (DEF_VERSION < 1.4);
    
    std::list<char*> files; 
    
    for(unsigned int i=1; i<argc; i++) {
        std::string par = std::string(argv[i]);
        if((par.length() > 1) && par[0] == '-') {
            if (par[1] == 'w' && (i<(argc+1))) {
                windowName = argv[i+1];
                i++;
            } else if (par[1] == 't' && (i<(argc+1))) {
                interval = std::atoi(argv[i+1]);
                i++;
                if (!interval) {
                    printUsage();
                    return 0;
                }
            } else if (par[1] == 's' && (i<(argc+1))) {
                startupTime = atoi(argv[i+1]);
                i++;
                if (!startupTime) {
                    printUsage();
                    return 0;
                }
            } else if (std::string(par) == "--old") {
                isOld = true;
            }
        } else {
            files.push_back(argv[i]);
        }
    }
    if (!files.size()) {
        printUsage();
        return 0;
    }
    if(!gotoMinecraft(windowName, startupTime)) {
        std::cerr << "Couldn't find the " << windowName << " window!" << std::endl;
        return 0;
    } else {
        std::cout << "Found " << windowName << " window." << std::endl;
    }
    
    for(std::list<char*>::iterator it = files.begin(); it != files.end(); it++) {
        writeFile(*it, interval, isOld);
    }
    
    std::cout << "Done writing files." << std::endl;
    return EXIT_SUCCESS;
}
